/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Strae
 * D-76571 Gaggenau
 * T. +49 (0)7225 / 64-0
 * F. +49 (0)7225 / 64-300 
 */

package de.bsvrz.vew.sim.sim.simulation;

import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.TextValue;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.sim.sim.EmpfaengerMitAnAbmeldung;
import de.bsvrz.vew.sim.sim.SimulationApp;
import de.bsvrz.vew.sim.sim.sender.Sender;
import de.bsvrz.vew.sim.sim.simulationsstrecke.SimulationsStrecke;
import de.bsvrz.vew.sim.sim.steuerung.SteuerungSimulation;

/**
 * Eine Simulation als Empfnger und Sender fr Datenverteiler-Datenstze.
 * Diese abstrake Oberklasse kapselt diejenigen Funktionalitten, die unabhngig 
 * etwa von den Online -oder Offline Eigenschaften einer Simulation sind
 * *  
 * @version  $Revision: 1.3 $ / $Date: 2008/10/31 13:30:31 $ / ($Author: Sans $)
 * @author  Dambach-Werke GmbH
 * @author  Stefan Sans
 */
public abstract class Simulation implements EmpfaengerMitAnAbmeldung
{
  /**
   * Zustandsbezeichner
   */
  public enum ZustandsBezeichner
  {
    neu(0, "neu"),
    Vorstart(1, "Vorstart"),
    Start(2, "Start"),
    Stop(3, "Stop"),
    geloescht(4, "gelscht"),
    Pause(5, "Pause");
        
    public final int id;
    public final String text;
    
    public String toString()
    {
      return text;
    }
    
    private ZustandsBezeichner(int id, String text)
    {
      this.id = id;
      this.text = text;
    }
  }

  /**
   * Konfigurationsstatus bereit
   */
  public static final int KONFIG_STATUS_BEREIT = 0;
  
  /**
   * Konfigurationsstatus nicht bereit
   */
  public static final int KONFIG_STATUS_NICHT_BEREIT = 1;
  
  /**
   * Konfigurationsstatus gelscht
   */
  public static final int KONFIG_STATUS_GELOESCHT = 2;
  
  /**
   * Parametrierungsstatus aktiv
   */
  public static final int PARAM_STATUS_AKTIV = 0;
  
  /**
   * Parametrierungsstatus nicht aktiv
   */
  public static final int PARAM_STATUS_NICHT_AKTIV = 1;
  
  /**
   * Archivstatus aktiv
   */
  public static final int ARC_STATUS_AKTIV = 0;
  
  /**
   * Archivstatus nicht aktiv
   */
  public static final int ARC_STATUS_NICHT_AKTIV = 1;
  
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
   
  /**
   * Erzeugt eine neue Instanz vom Typ Simulation
   *  
   * @param obj das dynamische Simulationsobjekt
   * @param sv die Simulationsvariante
   * @param ss die Simulationsstrecke (Klasse Simulationsstrecke)
   * @param con die Datenverteiler-Verbindung
   */
  protected Simulation(SystemObject obj, short sv, 
      SimulationsStrecke ss, ClientDavInterface con)
  {
    _objekt = obj;
    _connection = con;
    
    _name = obj.getName();
    
    _simulationsVariante = sv;
    _simulationsStrecke = ss;
    
  }

  /**
   * Bestimmt die Steuerungsinstanz
   * @return die Steuerungsinstanz fr diese Simulation
   */
  public abstract SteuerungSimulation getSteuerung();
  
  /**
   * Bestimmt den endlichen Automaten
   * @return den endlichen Automaten fr diese Instanz
   */
  protected abstract EndlicherAutomat getEndlicherAutomat();
  
  /**
   * Der Name dieser Simulation
   */
  private String _name;
  
  /**
   * Die Simulationsvariante fr diese Simulation
   */
  private short _simulationsVariante;
  
  /**
   * Die dieser Simuolation zugeordnete Simuationsstrecke
   */
  private SimulationsStrecke _simulationsStrecke;
  
  /**
   * Das dynamische Simulations-Objekt
   */
  private SystemObject _objekt;
  
  /**
   * Die Datenverteiler-Verbindung
   */
  private ClientDavInterface _connection;
  
  private int _statusKonfiguration = KONFIG_STATUS_NICHT_BEREIT;
  private int _statusParametrierung = PARAM_STATUS_NICHT_AKTIV;
  private int _statusArchiv = ARC_STATUS_NICHT_AKTIV;
  
  /**
   * Senderinstanz, ber die das Senden als Quelle abgewickelt wird
   */
  private Sender _quelle;
  
  /**
   * Menge der Sendeanmeldungen, die beim bereinigen dieser Simulation 
   * abgemeldet werden mssen
   */
  private Set<DataDescription> _senderAnmeldungen;

 /**
  * Sende die bergebenen Daten mit dem zugeordneten  dynamischen 
  * Simuloations-Objekt als Quelle
  *  
  * @param atg die Attributgruppe
  * @param asp der Aspekt
  * @param data der Datensatz
  * @return true, wenn die Date erfolgreich versendet wurden, false sonst
  */
  public boolean sendeAlsQuelle(AttributeGroup atg, Aspect asp, Data data)
  {
    if (_quelle == null)
    {
      _quelle = new Sender(getConnection(), true);
    }
    
    short sv = getVarianteFuerSimulation();
    boolean result = _quelle.send(getObjekt(), atg, asp, sv , data);
    
    if (_senderAnmeldungen == null)
    {
      _senderAnmeldungen = new HashSet<DataDescription>();
    }
    _senderAnmeldungen.add(new DataDescription(atg, asp, sv));
    
    return result;
  }
  
  /**
   * Bestimmt den Namen dieser Simulation
   * 
   * @return den Namen der Simulation
   */
  public String getName()
  {
    return _name;
  }
  
  @Override
  public String toString()
  {
    return _name;
  }
  
  /**
   * Bestimmt die Simulationsvariante dieser Simulation
   * 
   * @return die Simulationsvariante der Simulation
   */
  public short getSimulationsVariante()
  {
    return _simulationsVariante;
  }
   
  /**
   * Bestimmt die Simulationsvariante unter der diese 
   * Simulation Verwaltungsinformationen versendet und empfngt
   * 
   * @return die beschriebene Simulationsvariante
   */
  public short getVarianteFuerSimulation()
  {
    return 0; 
    // return getSimulationsVariante();
  }
  
  /**
   * Bestimmt die Simulationsstrecke fr dieser Simulation
   * @return die Simulationsstrecke der Simulation
   */
  public SimulationsStrecke getSimulationsStrecke()
  {
    return _simulationsStrecke;
  }

  public int getStatusKonfiguration()
  {
    return _statusKonfiguration;
  }
  
  public int getStatusParametrierung()
  {
    return _statusParametrierung;
  }
  
  public int getStatusArchiv()
  {
    return _statusArchiv;
  }
  
  /**
   * Bestimmt den aktuellen Zustand dieser Simulation
   * 
   * @return den aktuellen Zustand der Simulation
   */
  public Zustand getZustand()
  {
    // TODO Auto-generated method stub
    return getEndlicherAutomat().getMomentZustand();
  }
  
  /**
   * Setzt den aktuellen Zustand dieser Simulation
   * 
   * @param uebergang die Id des Zustand-bergangs
   * @return true, wenn der Zustandsbergang erfolgreich 
   * vollzogen wurde, false sonst
   */
  public boolean setZustand(int uebergang)
  {
    if (getSteuerung().erstelleWorkflow(uebergang))
    { 
      try
      {
        getEndlicherAutomat().realisiereUebergang(uebergang);
        return true;
      }  
      catch (IllegalStateException e)
      {
        _debug.error(e.getMessage());
      }    
    }
    return false;
  }
  
  /**
   * Setzt den aktuellen Zustand dieser Simulation
   * 
   * @param zustandNeu Name des Endzustands
   */
  public void setZustand(String zustandNeu)
  {
    Zustand[] zustaende = getEndlicherAutomat()._zustaende;
    Zustand vorgabe = null;
    
    for (Zustand zustand: zustaende)
    {
      if (zustand != null && zustand.getName().equals(zustandNeu))
      {
        vorgabe = zustand;
        break;
      }
    }
    
    if (vorgabe != null)
    {  
      try
      {
        int uebergang = getEndlicherAutomat().berechneUebergang(getZustand(), vorgabe);
        /*
        if (setZustand(uebergang))
        {
          getSteuerung().sendeZustand();
        }
        else
        {
          _debug.error("Simulation " + this + ": " +
              "bergang konnte nicht realisiert werden vom momentanen Zustand \"" +
              getZustand() + "\" in den Zustand \"" + vorgabe +"\"");
        }
        */
        if (!setZustand(uebergang))
        {
          _debug.error("Simulation " + this + ": " +
              "bergang konnte nicht realisiert werden vom momentanen Zustand \"" +
              getZustand() + "\" in den Zustand \"" + vorgabe +"\"");
        }
      }
      catch (UnsupportedOperationException e)
      {
        _debug.warning("Simulation " + this + ": " +
            "bergang nicht untersttzt vom momentanen Zustand \"" +
            getZustand() + "\" in den Zustand \"" + vorgabe +"\"");
      }
    }
    else
    {
      _debug.error("Vorgabe-Zustand \"" + zustandNeu +
          "\" nicht gefunden fr Simulation " + this);
    }  
    getSteuerung().sendeZustand();
  }
  
  /**
   * Setzt den aktuellen Zustand dieser Simulation asynchron
   * 
   * @param zustandNeu Name des Endzustands
   */
  public void starteZustandswechsel(final String zustandNeu)
  {
    Thread arbeiter = new Thread()
    {
      public void run()
      {
        synchronized (Simulation.this)
        {
          setZustand(zustandNeu);
        }
      }
    };
    arbeiter.start();
  }  

  /**
   * Realisiert den automatsichen Wechsel in den Zustand "Stop" asynchron.
   * Dies ist dann ntig, wenn bei einer laufenden Simulation die Konfiguration, 
   * die Parametrierung oder das Archiv nicht mehr bereit sind 
   */
  private void starteWechselNachStop()
  {
    synchronized(this)
    {
      Zustand zustand = getZustand();
      if (zustand.getName().equals(ZustandsBezeichner.Start.text) || 
          zustand.getName().equals(ZustandsBezeichner.Pause.text))
      {
        starteZustandswechsel(ZustandsBezeichner.Stop.text);
      }
    }  
  }
  
  public SystemObject getObjekt()
  {
    return _objekt;
  }
  
  public ClientDavInterface getConnection()
  {
    return _connection;
  }

  public void anmelden()
  {
    if (getObjekt().isOfType("typ.simulation"))
    {
      short simulationsVariante = getVarianteFuerSimulation();
      ReceiveOptions anmeldeoptionen = ReceiveOptions.normal();
      
      Aspect aspekt = getObjekt().getDataModel().getAspect("asp.zustand");
      
      ReceiverRole empfaengerrolle = SimulationApp.isArchivVorV3() ?
          ReceiverRole.drain() : ReceiverRole.receiver();
      
      AttributeGroup attributgruppe = getObjekt().getDataModel().getAttributeGroup("atg.simulationsStatusArchiv");
      DataDescription datenbeschreibung = new DataDescription(attributgruppe, aspekt, simulationsVariante);
      getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung, anmeldeoptionen, empfaengerrolle);

      empfaengerrolle = ReceiverRole.receiver();
      
      attributgruppe = getObjekt().getDataModel().getAttributeGroup("atg.simulationsStatusParametrierung");
      datenbeschreibung = new DataDescription(attributgruppe, aspekt, simulationsVariante);
      getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung, anmeldeoptionen, empfaengerrolle);
      
      attributgruppe = getObjekt().getDataModel().getAttributeGroup("atg.simulationsStatusKonfiguration");
      datenbeschreibung = new DataDescription(attributgruppe, aspekt, simulationsVariante);
      getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung, anmeldeoptionen, empfaengerrolle);
    }
  }
  
  public void abmelden()
  {
    try
    {
      if (getObjekt().isOfType("typ.simulation"))
      {
        short simulationsVariante = getVarianteFuerSimulation();
        Aspect aspekt = getObjekt().getDataModel().getAspect("asp.zustand");
        AttributeGroup attributgruppe = getObjekt().getDataModel().getAttributeGroup("atg.simulationsStatusArchiv");
        DataDescription datenbeschreibung = new DataDescription(attributgruppe, aspekt, simulationsVariante);
        getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);
         
        attributgruppe = getObjekt().getDataModel().getAttributeGroup("atg.simulationsStatusParametrierung");
        datenbeschreibung = new DataDescription(attributgruppe, aspekt, simulationsVariante);
        getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);
        
        attributgruppe = getObjekt().getDataModel().getAttributeGroup("atg.simulationsStatusKonfiguration");
        datenbeschreibung = new DataDescription(attributgruppe, aspekt, simulationsVariante);
        getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);
        
        if (_quelle != null && _senderAnmeldungen != null)
        {  
          for (DataDescription dd: _senderAnmeldungen)
          {
            getConnection().unsubscribeSender(_quelle, getObjekt(), dd);
          }
        }  
      }
    }
    catch (IllegalStateException e)
    {
      _debug.error("Fehler beim Abmelden der Empfngeranmeldungen von Simulation " + getObjekt(), e);
    }
  }
    
  public void update(ResultData[] resultData)
  {
    for (ResultData data : resultData)
    {
      _debug.finest("Daten empfangen:\n" + data);
      if (data.getDataState() == DataState.DATA)
      {
        Aspect asp = data.getDataDescription().getAspect();
        AttributeGroup atg = data.getDataDescription().getAttributeGroup();
        short sv = data.getDataDescription().getSimulationVariant();
        
        _debug.finer("Daten empfangen mit atg/asp/sv: " + atg.getPid() + "/" +
            asp.getPid() + "/" + sv + ":\n" + data);
        
        if (sv == getVarianteFuerSimulation() && asp.getPid().equals("asp.zustand"))
        {
          if (atg.getPid().equals("atg.simulationsStatusKonfiguration"))
          {
            _debug.fine("Bearbeite Daten mit atg/asp/sv: " + atg.getPid() + "/" +
                asp.getPid() + "/" + sv + ":\n" + data);
            
            TextValue textValue = data.getData().getItem("KonfigurationZustand").asTextValue();
            String zustand = textValue.getText();
            _debug.config("Konfiguration meldet \"KonfigurationZustand: " + 
                zustand + "\" fr Simulation " + this);
            
            if (zustand.equals("bereit"))
            {
              _statusKonfiguration = KONFIG_STATUS_BEREIT;
            }
            else if (zustand.equals("gelscht"))
            {
              _statusKonfiguration = KONFIG_STATUS_GELOESCHT;
            }
            else
            {
              _statusKonfiguration = KONFIG_STATUS_NICHT_BEREIT;
            }
            if (_statusKonfiguration != KONFIG_STATUS_BEREIT)
            {
              starteWechselNachStop();
            }
          }
          if (atg.getPid().equals("atg.simulationsStatusParametrierung"))
          {
            _debug.fine("Bearbeite Daten mit atg/asp/sv: " + atg.getPid() + "/" +
                asp.getPid() + "/" + sv + ":\n" + data);
            
            TextValue textValue = data.getData().getItem("ParametrierungBereit").asTextValue();
            String zustand = textValue.getText();
            _debug.config("Parametrierung meldet \"ParametrierungBereit: " + 
                zustand + "\" fr Simulation " + this);
            
            if (zustand.equals("Ja"))
            {
              _statusParametrierung = PARAM_STATUS_AKTIV;
            }
            else
            {
              _statusParametrierung = PARAM_STATUS_NICHT_AKTIV;
            }
            if (_statusParametrierung != PARAM_STATUS_AKTIV)
            {
              starteWechselNachStop();
            }
          }
          if (atg.getPid().equals("atg.simulationsStatusArchiv"))
          {
            _debug.fine("Bearbeite Daten mit atg/asp/sv: " + atg.getPid() + "/" +
                asp.getPid() + "/" + sv + ":\n" + data);
            
            TextValue textValue = data.getData().getItem("ArchivBereit").asTextValue();
            String zustand = textValue.getText();
            _debug.config("Archiv meldet \"ArchivBereit: " + 
                zustand + "\" fr Simulation " + this);
            if (zustand.equals("Ja"))
            {
              _statusArchiv = ARC_STATUS_AKTIV;
            }
            else
            {
              _statusArchiv = ARC_STATUS_NICHT_AKTIV;
            }
            if (_statusArchiv != ARC_STATUS_AKTIV)
            {
              starteWechselNachStop();
            }
          }
        }
      }  
    }  
  }
  
}
